﻿namespace HeuristicLab.Algorithms.PushGP.Expressions {

  using HeuristicLab.Algorithms.PushGP.Interpreter;
  using HeuristicLab.Algorithms.PushGP.Stack;

  public class LoopState {
    public Expression Body;
    public long CurrentIndex;
    public long DestinationIndex;
    public long Incrementor;
  }

  public abstract class LoopExpression : StatefullExpression<LoopState> {
    protected LoopExpression(string stringRepresentation) : base(new LoopState(), stringRepresentation) { }

    protected void Eval(IPushGpInterpreter interpreter, IStack<Expression> sourceStack, bool pushCurrentIndex = false) {
      // if not initialized
      if (this.State.Body == null) {
        if (this.HasInsufficientArguments(interpreter, sourceStack)) return;

        this.InitState(interpreter, sourceStack);

        interpreter.ExecStack.Push(this, this.State.Body);

        return;
      }

      // if loop end reached
      if (this.State.DestinationIndex == this.State.CurrentIndex) {
        this.PushLastIteration(interpreter);
        this.Clear();

        return;
      }

      this.PushIteration(interpreter);
      this.State.CurrentIndex += this.State.Incrementor;
    }

    protected virtual void Clear() {
      this.State.Body = null;
    }

    protected virtual void PushIteration(IPushGpInterpreter interpreter) {
      interpreter.IntegerStack.Push(this.State.CurrentIndex);
      interpreter.ExecStack.Push(this, this.State.Body);
    }

    protected virtual void PushLastIteration(IPushGpInterpreter interpreter) {
      interpreter.IntegerStack.Push(this.State.CurrentIndex);
      interpreter.ExecStack.Push(this.State.Body);
    }

    public override bool Equals(object obj) {
      if (this == obj)
        return true;

      if (obj == null || obj.GetType() != this.GetType())
        return false;

      var other = obj as LoopExpression;

      if (this.State.CurrentIndex != other.State.CurrentIndex ||
          this.State.DestinationIndex != other.State.DestinationIndex ||
          this.State.Incrementor != other.State.Incrementor)
        return false;

      var isBodyEqual = this.State.Body == other.State.Body ||
                        (this.State.Body != null &&
                        other.State.Body != null &&
                        this.State.Body.Equals(other.State.Body));

      return isBodyEqual;
    }

    protected override int CalcHashCode() {
      var hash = 19;

      if (this.State.Body != null)
        hash = hash * 31 + this.State.Body.GetHashCode();

      hash = hash * 31 + this.State.CurrentIndex.GetHashCode();
      hash = hash * 31 + this.State.DestinationIndex.GetHashCode();
      hash = hash * 31 + this.State.Incrementor.GetHashCode();

      return hash;
    }

    protected abstract bool HasInsufficientArguments(IPushGpInterpreter interpreter, IStack<Expression> sourceStack);

    protected abstract void InitState(IPushGpInterpreter interpreter, IStack<Expression> sourceStack);
  }
}